/*
 * ADAPT2D : a software for automatic mesh adaptation in 2D
 *
 * AUTHOR : Manuel J. Castro Diaz(e-mail:castro@gamba.cie.uma.es)
 * ADAPTED FOR FREEFEM : Prud'homme Christophe (e-mail:prudhomm@ann.jussieu.fr) 
 *
 * this code is public domain
 * 
 * You may copy freely these files and use it for    
 * teaching or research. These or part of these may   
 * not be sold or used for a commercial purpose without
 * our consent
 * 
 * Any problems should be reported to the AUTHOR
 * at the following address : castro@gamba.cie.uma.es
 */


#include <header.hxx>
#include <t_t1_dl.hxx>
#include <a_t1_dl.hxx>
#include <v_t1_dl.hxx>

void verifica (Triangulo_T1_dlist* Ltr, Arista_T1_dlist* Lar,\
               Vertice_T1_dlist* Lvr) {


  Triangulo_T1_dlink* Lt_lk;
  Arista_T1_dlink* La_lk;
  Vertice_T1_dlink* Lv_lk;
  
  Triangulo_T1* t0,*t1,*t2;
  Arista_T1* a0;
  Vertice_T1* v0,*s0,*s1;
  Scalar temp;
  int err=0;
  int pos=0;
  int pos0,pos1;
  int i;
#ifdef DEBUG
  std::cout<<"Mesh verification:"<<std::endl;
  std::cout<<"-----------------"<<std::endl;
  std::cout<<"Number of vertices:"<<(Lvr->num_elem())<<std::endl;
  std::cout<<"Vertex verification:"<<std::endl;
#endif /* DEBUG */
  
  Lv_lk=Lvr->principio();
  while (Lv_lk) {
    v0=Lv_lk->s;
    if (v0->lk!=Lv_lk) {
      std::cout<<"Error in vertex-link:"<<std::endl;
      std::cout<<*v0<<std::endl<<"Link:"<<Lv_lk<<std::endl;
      err++;
    }
    t0=v0->t;
    if (t0==NIL) {
      err++;
      std::cout<<"Triangle pointed by NIL vertex"<<std::endl;
      std::cout<<*v0<<std::endl<<t0<<std::endl;
    }
    else {
      pos=t0->sommet0(v0);
      if (pos==-1) {
        err++;
        std::cout<<"Error vertex-Triangle:"<<std::endl;
        std::cout<<*v0<<std::endl<<*t0<<std::endl;
      }
    }
    Lv_lk=Lv_lk->sig();
  }
#ifdef DEBUG
  std::cout<<"End testing vertices:"<<std::endl;
  std::cout<<std::endl<<"----------------------------------------------------"
      <<std::endl;
  std::cout<<"Number of edges:"<<(Lar->num_elem())<<std::endl;
  std::cout<<"Edge verification:"<<std::endl;
#endif /* DEBUG */
  La_lk=Lar->principio();
  while (La_lk) {
    a0=La_lk->a;
    if (a0->lk!=La_lk) {
      std::cout<<"Error edge-link:"<<std::endl;
      std::cout<<*a0<<std::endl<<"Link:"<<La_lk<<std::endl;
      err++;
    }
    s0=a0->s[0];
    s1=a0->s[1];
    t0=a0->tr;
    if (t0==NIL) {
      err++;
      std::cout<<"Error, Triangle pointed NIL:"<<std::endl;
      std::cout<<*a0<<std::endl<<t0<<std::endl;
    }
    else {
      if (s0==NIL || s1==NIL || s0==s1) {
        std::cout<<"Error, vertices NIL or identical:"<<std::endl;
        std::cout<<*a0<<std::endl<<s0<<"    "<<s1<<std::endl;
        err++;
      }
      else {
        pos0=t0->arista0(s0,s1);
        pos1=t0->arista0(a0);
        if (pos0==-1 && pos1==-1) {
          err++;
          std::cout<<"edge does'nt exist in triangle:"<<std::endl;
          std::cout<<*a0<<std::endl<<*t0<<std::endl;
        }
        if (pos0==-1 && pos1!=-1) {
          err++;
          std::cout<<"vertices don't appear:"<<std::endl;
          std::cout<<*a0<<std::endl<<*s0<<std::endl<<*s1<<std::endl<<*t0<<std::endl;
        }
        if (pos1==-1 && pos0!=-1) {
          err++;
          std::cout<<"edge does'nt appear:"<<std::endl;
          std::cout<<*a0<<std::endl<<*s0<<std::endl<<*s1<<std::endl<<*t0<<std::endl;
        }
        if (pos1!=-1 && pos0!=-1 && pos0!=pos1) {
          err++;
          std::cout<<"Edge not in the same place."<<std::endl;
          std::cout<<*a0<<std::endl<<*s0<<std::endl<<*s1<<std::endl<<*t0<<std::endl;
        }
        t1=t0->t[pos0];
        if (t1) {
          pos0=t1->arista0(s0,s1);
          pos1=t1->arista(a0);
          if (pos0==-1 && pos1==-1) {
            err++;
            std::cout<<"edge doesn't exist in triangle:"<<std::endl;
            std::cout<<*a0<<std::endl<<*t1<<std::endl;
          }
          if (pos0==-1 && pos1!=-1) {
            err++;
            std::cout<<"vertices don't appear:"<<std::endl;
            std::cout<<*a0<<std::endl<<*s0<<std::endl<<*s1<<std::endl<<*t1<<std::endl;
          }
          if (pos1==-1 && pos0!=-1) {
            err++;
            std::cout<<"edege doesn't appear:"<<std::endl;
            std::cout<<*a0<<std::endl<<*s0<<std::endl<<*s1<<std::endl<<*t1<<std::endl;
          }
          if (pos1!=-1 && pos0!=-1 && pos0!=pos1) {
            err++;
            std::cout<<"edge not in the same place."<<std::endl;
            std::cout<<*a0<<std::endl<<*s0<<std::endl<<*s1<<std::endl<<*t1<<std::endl;
          }
        }
      }
    }
    La_lk=La_lk->sig();
  }     
#ifdef DEBUG
  std::cout<<"End edge verification."<<std::endl;
  std::cout<<std::endl<<"----------------------------------------------------"
      <<std::endl;
  std::cout<<"Number of triangles:"<<(Ltr->num_elem())<<std::endl;
  std::cout<<"Triangle verification:"<<std::endl;
#endif /* DEBUG */
  Lt_lk=Ltr->principio();
  while (Lt_lk) {
    t0=Lt_lk->t;
    temp=t0->area2D();
    if (temp<1e-12) {
      std::cout<<"Error triangle area:"<<std::endl;
      std::cout<<*t0<<std::endl;
      err++;
    }
    if (t0->lk!=Lt_lk) {
      std::cout<<"Error triangle-link:"<<std::endl;
      std::cout<<*t0<<std::endl<<"Link:"<<Lt_lk<<std::endl;
      err++;
    }
    for (i=0; i<3; i++) {
      v0=t0->s[i];
      if (v0==NIL) {
        std::cout<<"Error triangle-vertex NIL:"<<std::endl;
        std::cout<<*t0<<std::endl<<v0<<std::endl;
        err++;
      }
    }
    for (i=0; i<3; i++) { 
      a0=t0->a[i];
      t1=t0->t[i];
      if (a0==NIL) {
        std::cout<<"Error triangle-edge NIL:"<<std::endl;
        std::cout<<*t0<<std::endl<<a0<<std::endl;
        err++;
      }
      else {
        t2=a0->tr;
        if (t1) {
          if (t2!=t1 && t2!=t0) {
            err++;
            std::cout<<"Error edge-triangle not equal:"<<std::endl;
            std::cout<<*t0<<std::endl<<*t1<<std::endl<<*a0<<std::endl;
          }
        }  
        else {
          if (t2!=t0) {
            err++;
            std::cout<<"Error edge-triangle not equal:"<<std::endl;
            std::cout<<*t0<<std::endl<<*t1<<std::endl<<*a0<<std::endl;
          }
        }
        s0=a0->s[0];
        s1=a0->s[0];
        pos=t0->arista0(a0);
        if (pos!=i) {
          err++;
          std::cout<<"Error edge-vertices-triangle not equal:"<<std::endl;
          std::cout<<*t0<<std::endl<<*a0<<std::endl<<*s0<<std::endl<<*s1<<std::endl;
        }
      }
    }
    for (i=0; i<3; i++) {
      t1=t0->t[i];
      if (t1) {
        a0=t0->a[i];
        pos=t1->arista0(a0);
        if (pos==-1) {
          err++;
          std::cout<<"Error triangle-neighboor_triangle:"<<std::endl;
          std::cout<<*t0<<std::endl<<*t1<<std::endl<<*a0<<std::endl;
        }
      }
    }
    Lt_lk=Lt_lk->sig();
  }
#ifdef DEBUG
  std::cout <<"End of verification:"<<std::endl;
  std::cout <<"Number of errors:"<<err<<std::endl;
  std::cout <<"--------------------------------------------"<<std::endl;
#endif /* DEBUG */
  if (err!=0) {
    std::cerr<<" Sorry too many errors in verifica."<<std::endl;
    exit(1);
  }
}
