//
// SUMMARY:      Xrgraph.c graphics under X11
// USAGE:        
//
// AUTHOR:       Christophe Prud'Homme
// ORG:          Numerica
// ORIG-DATE:     10-Aug-95
// LAST-MOD:      8-Feb-97 at 14:22:15 by Christophe Prud'homme
// DESCRIPTION: routines for X11 and workstations + system dependent routines
// DESCRIP-END.
//

#define _sys_time_h
// C include
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>

// c++ include
//#include <new.h>
#include <iostream>
#include <opclass.h>
#include <rgraph.h>
#include <disk.h>
#include <lexical.h>
#include <syntaxic.h>

#ifdef HPPA
#ifndef __GNUC__
typedef char *caddr_t;
#endif
#endif

#if 1	// K.H
#include <signal.h>
#endif

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>


/*
 * this is for HP workstations
 * this allow to avoid including setjmp.h
 *
 * SEE THE FUNCTION ERREUR (INT)
 */
#if ( defined (__hp9000s700) )
# ifdef __cplusplus
extern "C" {
# endif /* __cplusplus */
# define _JBLEN 50
  typedef double jmp_buf[_JBLEN/2];
  extern void longjmp (jmp_buf,int);
  extern int setjmp (jmp_buf);
# ifdef __cplusplus
}
# endif /* __cplusplus */
#else /* not __hp9000s700 */
# include <setjmp.h>
#endif /* defined (__hp9000s700) */
 

#define reel float 

typedef enum {
  NO_MESSAGE=-2,
  STOP = -1,
  ERROR = 0,
  MESH = 1,
  FUNCTION = 2
} compil_msg;
compil_msg cmsg;

#if defined(XGFEM)
# include <comm.hxx>
extern comm slave;
void SlaveDeconnect ();
void SlaveSendError (int&, char *, int&);
#endif /* XGFEM */

jmp_buf context;
extern void parse (char *); 
extern int NumOfIterations,Iter;
extern triangulation t;
extern float xfmin, xfmax;


static  Display *display;
static  Window win;
static  XSizeHints size_hints;
static  GC gc;
static  XFontStruct *font_info;
static reel echx,echy,rxmin,rxmax,rymin,rymax;
static int screen, width, height, currx, curry;
#if 1	// K.H
static unsigned long color[16];
static char *cmd;
static char tmp_file[64];
#include <unistd.h>

#define TMPDIR	"/var/tmp"
#endif

void
out_of_memory ()
{
  std::cerr << "FreeFEM error: operator new failed; not enough memory" << std::endl;
#if defined(XGFEM)
  SlaveDeconnect ();
#endif /* XGFEM */
  exit (-1);
}
void
NEW_HANDLER (void)
{
  std::set_new_handler (&out_of_memory);
}
void message(char *s)
{
  printf("%s \n",s);
}

void erreur(char *s)
{
  extern jmp_buf context;
  message(s);
  cmsg = ERROR;
#if defined(XGFEM)
  SlaveSendError ((int&)cmsg, s, numligne);
  SlaveDeconnect ();
#endif /* XGFEM */
#if 1
  if (strcmp(s, " interpretor stopped") == 0)
	exit(0);
#endif
  longjmp (context, -1);
}

void rflush()
{
  XFlush(display);
}

void rattente(int waitm)
{
  char click[] = "Click to continue...";
  char values[256];
  XEvent report;

#if 1	// K.H
  couleur(0);
#endif
  if (xfmin != 0.0 && xfmax != 0.0)
    {
      sprintf(values, "Mini/Maxi: [ %f , %f ]", xfmin, xfmax);
      XDrawString (display,
                   win,
                   gc,
                   5,40,
                   values,
                   strlen(values));
    }
  if (NumOfIterations > 0)
    {
      sprintf(values, "Iter: %d", Iter);
      XDrawString (display,
                   win,
                   gc,
                   5,60,
                   values,
                   strlen(values));
    }
  if (waitm)
    {
      XDrawString (display,
                   win,
                   gc,
                   5,20,
                   click,
                   strlen(click));
      do XNextEvent(display, &report);
      while (report.type != ButtonPress && report.type != KeyPress);
    }
  XCheckMaskEvent(display, ButtonPressMask,&report);
  if (report.type == ButtonPress)
    if (report.xbutton.button == Button3)
      {
        extern jmp_buf context;
//        type_msg = STOP;
        cmsg = STOP;
#if defined(XGFEM)
        slave.csend ((int)cmsg);
        cmsg = NO_MESSAGE;
        slave.csend ((int)cmsg);
        slave.cclose(1);
#endif /* XGFEM */
        closegraphique();
        longjmp (context, -1);
      }
  XFlush (display);
}

int xerror()
{
  fprintf(stderr, "Probleme avec X-Windows\n");
  return 1;
}

void initgraphique()
{
#if 1	// K.H
  static char *colorname[] = {
	(char *)"black",
	(char *)"slate blue",
	(char *)"red",
	(char *)"coral",
	(char *)"blue violet",
	(char *)"yellow",
	(char *)"lime green",
	(char *)"spring green",
	(char *)"khaki",
	(char *)"light gray",
	(char *)"white",
	(char *)"blue",
	(char *)"magenta",
	(char *)"green",
	(char *)"cyan",
	(char *)"orange"
  };
  Colormap colmap;
  XColor c0, c1;
  int i;
#endif
  XEvent report;
  XSetWindowAttributes attrs;
  unsigned long attrs_mask;

  display = XOpenDisplay(NULL);

  if ((font_info = XLoadQueryFont(display, "7x13")) == NULL)
   {
    fprintf (stderr, "FreeFEM: cannot open 7x13 font\n");
    fprintf (stderr, "FreeFEM: I am going to try an other one.\n");
    if ((font_info = XLoadQueryFont(display, "9x15")) == NULL)
	{
	  fprintf (stderr, "FreeFEM: cannot open 9x15 font\n");
          fprintf (stderr, "FreeFEM: AArgghh no misc font.\n");
        }
   }

  XSetErrorHandler((XErrorHandler)xerror);
  XSetIOErrorHandler((XIOErrorHandler)xerror);
  screen = DefaultScreen(display);
  width = DisplayWidth(display, screen) - 100;
  height = DisplayHeight(display, screen) - 160;
  win = XCreateSimpleWindow(display, RootWindow(display, screen), 50, 80,
                            width, height, 4, 
                            BlackPixel(display, screen), WhitePixel(display, screen));

  size_hints.flags = PPosition | PSize;
  size_hints.x = 0;
  size_hints.y = 0;
  size_hints.width = width;
  size_hints.height = height;

  XSetStandardProperties(display, win, "ploth", NULL, 0, NULL, 0, &size_hints);
  XSelectInput(display, win, ExposureMask | ButtonPressMask);
  gc = XCreateGC(display, win, 0, NULL);
  XSetFont(display, gc, font_info->fid);
  XSetForeground(display, gc, BlackPixel(display, screen)); 
  XMapWindow(display, win);

  //
  // enable backing store
  // WARNING: not all display are able to do backing 
  //
  attrs.backing_store = DoesBackingStore (ScreenOfDisplay(display,screen));
  attrs_mask = CWBackingStore;
  XChangeWindowAttributes (display, win, attrs_mask, &attrs);
#if 1	// K.H
  colmap = XDefaultColormap(display, 0);
  for (i = 0; i < 16; i++) {
	XAllocNamedColor(display, colmap, colorname[i], &c1, &c0);
	color[i] = c1.pixel;
  }
#endif
    
  do XNextEvent(display, &report); while (report.type != Expose);
}

void closegraphique()
{
  XUnloadFont(display, font_info->fid);
  XFreeGC(display, gc);
  XCloseDisplay(display);
}

void cadre(reel xmin,reel xmax,reel ymin,reel ymax)
{
  rxmin = xmin;
  rxmax = xmax;
  rymin = ymin;
  rymax = ymax;

  echx = width / (xmax - xmin);
  echy = height / (ymax - ymin);
}

void couleur(int c)
{
/*
  setcolor(lacouleur = c);
*/
#if 1	// K.H
  XSetForeground(display, gc, color[c & 0xff]);
#endif
}

int scalx(reel x)
{
  return (int)((x - rxmin) * echx);
}

int scaly(reel y)
{
  return (int)((rymax - y) * echy);
}

void pointe(reel x, reel y)
{
  XDrawPoint(display, win, gc, scalx(x), scaly(y));
}

void rmoveto(reel x, reel y)
{
  currx = scalx(x);
  curry = scaly(y);
}

void rlineto(reel x, reel y)
{
  int newx = scalx(x), newy = scaly(y);
  XDrawLine(display, win, gc, currx, curry, newx, newy);
  currx = newx; curry = newy;
/*   XFlush(display); */
}

#if 1	// K.H
void label(reel x, reel y, reel r, char *str)
{
  int x1, y1, n, wid;

  n = strlen(str);
  wid = n * 7;	/* 7x13 font */
  x1 = scalx(x) - wid /2;
  y1 = scaly(y) + 4;
#if 0
  XDrawImageString(display, win, gc, x1, y1, str, n);
#else
  XDrawString(display, win, gc, x1, y1, str, n);
#endif
}

void  format(char *fmt, reel p[4])
{
  int i = 0;
  char c, d, *s, *t, str[512];

  for (s = fmt, t = str; *s; ) {
	c = *t++ = *s++;
	if (c == '\\') {
		if ((c = *s++) == 'n')
			d = '\n';
		else if (c == 't')
			d = '\t';
		else if (c == 'r')
			d = '\r';
		else if (c == 'f')
			d = '\f';
		else if ('0' <= c && c <= '9') {
			for (d = c; '0' <= *s && *s <= '9'; s++)
				d = d * 8 + *s - '0';
		}
		else
			d = c;
		t[-1] = d;
		continue;
	}
	if (c != '%')
		continue;
	if (*s == '%') {
		*t++ = *s++; continue;
	}
	while (*s && (*s == 'l' || *s < 'a' || 'z' < *s))
		*t++ = *s++;
	d = *t++ = *s++;
	*t = '\0';
	switch (d) {
	case 'e': case 'f': case 'g':
		if (i<4)
			printf(str, p[i++]);
		else
			printf("?");
		break;
	case 'd':
		if (i<4)
			printf(str, (int)p[i++]);
		else
			printf("?");
		break;
	default:
		printf("%s", str);
		break;
	}
	t = str;
  }
  *t = '\0';
  printf("%s", str);
  fflush(stdout);
}

void xsleep(int n)
{
  usleep(n*1000);
}
#endif

void cadreortho(reel centrex, reel centrey, reel rayon)
{

  if (height < width)
  {
    rymin = centrey - rayon;
    rymax = centrey + rayon;
    echx = echy= height / (2 * rayon);
    rxmin= centrex - width / (2 * echx);
    rxmax= centrex + width / (2 * echx);
  }
  else
  {
    rxmin = centrex - rayon;
    rxmax = centrex + rayon;
    echx = echy = width / (2 * rayon);
    rymin = centrey - height / (2 * echy);
    rymax = centrey + height / (2 * echy);
  }
}

void cercle(reel centrex, reel centrey, reel rayon)
{
  int r = (int)(rayon * echx);
  XDrawArc(display, win, gc,
     scalx(centrex) - r, scaly(centrey) - r, width, height, 0, 360 * 64);
  XFlush(display);
}
void reffecran()
{
 XClearWindow(display,win);
}

void raffpoly(int n, float *poly)
{
  int i, *poly0;
  
  poly0=new int[2*n];
  if(poly0 != NULL)
  {
    fprintf(stderr, "Erreur d'allocation dans raffpoly\n");
    return;
  }
  for(i=0; i<n; i++)
  {
    poly0[2*i]=scalx(poly[2*i]);
    poly0[2*i+1]=scaly(poly[2*i+1]);
  }
/*  fillpoly(n, poly0); */
  delete [] poly0;
}

int getprog(char* fn,int argc, char **argv)
{
  if (argc != 2) return 0;
  strcpy(fn,argv[1]);
#if 0	// K.H
  printf(" file : %s\n",fn);
#endif
  return argc;
}

void execute (char * str)
{
  system(str);
}

/*
 * Main program:
 *   - read the file and store it in prog
 *   - parse prog using the routine parse (see syntaxic)
 */
int
main (int argc, char **argv)
{
  char       *prog;
  char       *fname;
#if 1	// K.H
  char	     *s;
  extern int show_ngt;
  extern int show_bdy;
  void onsig(int);
#endif
  extern char errbuf[];

  NEW_HANDLER (); // see dependent system files ({pc,x,mac}rgraph.{h,cpp})
  
  fname = new char[256];

#if !defined(XGFEM)
#if 1	// K.H
  cmd = argv[0];
  while (--argc > 0 && (*++argv)[0] == '-')
    for (s = argv[0]+1; *s; s++)
	switch (*s) {
	case 'r':
		show_ngt++;
		break;
	case 'b':
		show_bdy++;
		break;
	default:
		fprintf(stderr, "Usage: %s [-b -r] [file]\n", cmd);
		exit(1);
	}

  signal(SIGSEGV, onsig);
  signal(SIGFPE, onsig);

  if (argc == 0) {
	FILE *fp; int c;
	sprintf(tmp_file, "%s/FreeFEM%06d", TMPDIR, getpid());
	if ((fp = fopen(tmp_file, "w+")) == NULL) {
		fprintf(stderr, "can't open %s\n", tmp_file);
		exit(1);
	}
	while ((c = fgetc(stdin)) != EOF)
		fputc(c, fp);
	fclose(fp);
	strcpy(fname, tmp_file);
  }
  else if (argc == 1)
	strcpy(fname, argv[0]);
  argc = 2;
#else
	
  argc = getprog (fname, argc, argv);
#endif
  OPTION = 0;
  if (argc == 2)
    {
      prog = readprog (fname);
#if 1	// K.H
      if (tmp_file != NULL)
	unlink(tmp_file);
#endif
      delete [] fname;
      /*std::cerr << " PROGRAM:\n" 
	<< prog << "\n";*/
#endif /* not XGFEM */
      if (setjmp (context) == 0)
        {
          // parse prog buffer
          parse (prog); 
          delete [] prog;
          /* 
           * Freeing the triangulation
           */
          delete [] t.rp;
          delete [] t.tr;
          delete [] t.ng;
          delete [] t.ngt;
          
        }
      else
        {
          switch ((int)cmsg)
            {
            case ERROR:
              printf ("FreeFEM: an error has occured.\n");
              break;
            case STOP:
              printf ("FreeFEM: simulation stopped (right button pressed)\n");
              break;
            }
        }

#if !defined(XGFEM)
    }
  else
    {
      std::cerr << "Usage:\n"
           << "  freefem filename.pde\n";
    }
#endif /* XGFEM */

#if 0	// K.H
  std::cerr << "PASS: Test " << prog << " Ok";
#endif
  return 0;
}

#if 1	// K.H
void
onsig(int sig)
{
  if (sig == SIGSEGV)
	fprintf(stderr, "Segmentation fault.\n");
  else if (sig == SIGFPE)
	fprintf(stderr, "Floating point exception.\n");
  else
	fprintf(stderr, "Signal %d detected.\n", sig);
  exit(1);
}
#endif

#if defined(XGFEM)
void
SlaveConnect (char *&buffer, parameter& adapt_param)
{
  slave.init("slave", SOCK_STREAM,1);
  slave.cconnect();
  slave.crecv (buffer);
  slave.crecv (adapt_param);
}
void
SlaveDeconnect ()
{
  extern compil_msg cmsg;
  cmsg = NO_MESSAGE;
  slave.csend ((int)cmsg);
  slave.cclose(1);
}
void
SlaveSendMesh (triangulation& t)
{
  extern compil_msg cmsg;
  cmsg = MESH;
  int nb_p = t.np;
  int nb_t = t.nt;
  int i, total;

  slave.csend ((int)cmsg);
  slave.csend(nb_p);
  slave.csend(nb_t);
  
  for (i = 0;i < nb_p;i++)
    {
      slave.csend (t.rp[i]);
      slave.csend (t.ng[i]);
    }
  for (i = 0;i < nb_t;i++)
    {
      slave.csend (t.tr[i]);
      slave.csend (t.ngt[i]);
    }
}
void
SlaveSendFunction(int nquad, char *name, float *func)
{
  int i;
  
  extern compil_msg cmsg;
  cmsg = FUNCTION;
  slave.csend ((int)cmsg);
  slave.csend (nquad);
  slave.csend (name);
  for (i = 0;i < nquad;i++)
    slave.csend (func[i]);
}
void
SlaveSendError(int& msg, char *str, int& nl)
{
  slave.csend (cmsg);
  slave.csend (str);
  slave.csend (nl);
}
#endif /* XGFEM */
